#
# Copyright 2019 Bernhard Walter
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from jupyter_cadquery.cad_objects import _PartGroup, _Part, _show


class Part(_Part):
    def __init__(self, shape, name="part", color=None, show_faces=True, show_edges=True):
        super().__init__([shape], name, color, show_faces, show_edges)

    def to_assembly(self):
        return PartGroup([self])

    def show(self, grid=False, axes=False):
        return show(self, grid=grid, axes=axes)


class PartGroup(_PartGroup):
    def to_assembly(self):
        return self

    def show(self, grid=False, axes=False):
        return show(self, grid=grid, axes=axes)


class Assembly(PartGroup):
    def __init__(self, *args, **kwargs):
        import warnings

        super().__init__(*args, **kwargs)
        warnings.warn(
            "Class 'Assembly' is deprecated (too many assemblies ...). Please use class 'PartGroup' instead",
            RuntimeWarning,
        )


def show(cad_obj, **kwargs):
    """Show CAD objects in Jupyter

    Valid keywords:
    - height:            Height of the CAD view (default=600)
    - tree_width:        Width of navigation tree part of the view (default=250)
    - cad_width:         Width of CAD view part of the view (default=800)
    - bb_factor:         Scale bounding box to ensure compete rendering (default=1.5)
    - default_color:     Default mesh color (default=(232, 176, 36))
    - default_edgecolor: Default mesh color (default=(128, 128, 128))
    - render_edges:      Render edges  (default=True)
    - render_normals:    Render normals (default=False)
    - render_mates:      Render mates (for MAssemblies)
    - mate_scale:        Scale of rendered mates (for MAssemblies)
    - quality:           Linear deflection for tessellation (default=None)
                         If None, uses bounding box as in (xlen + ylen + zlen) / 300 * deviation)
    - deviation:         Deviation from default for linear deflection value ((default=0.1)
    - angular_tolerance: Angular deflection in radians for tessellation (default=0.2)
    - edge_accuracy:     Presicion of edge discretizaion (default=None)
                         If None, uses: quality / 100
    - optimal_bb:        Use optimal bounding box (default=False)
    - axes:              Show axes (default=False)
    - axes0:             Show axes at (0,0,0) (default=False)
    - grid:              Show grid (default=False)
    - ticks:             Hint for the number of ticks in both directions (default=10)
    - ortho:             Use orthographic projections (default=True)
    - ambient_intensity  Intensity of ambient ligth (default=1.0)
    - direct_intensity   Intensity of direct lights (default=0.12)
    - transparent:       Show objects transparent (default=False)
    - position:          Relative camera position that will be scaled (default=(1, 1, 1))
    - rotation:          z, y and y rotation angles to apply to position vector (default=(0, 0, 0))
    - zoom:              Zoom factor of view (default=2.5)
    - reset_camera:      Reset camera position, rotation and zoom to default (default=True)
    - mac_scrollbar:     Prettify scrollbars (default=True)
    - display:           Select display: "sidecar", "cell", "html"
    - tools:             Show the viewer tools like the object tree
    - timeit:            Show rendering times, levels = False, 0,1,2,3,4,5 (default=False)

    For example isometric projection can be achieved in two ways:
    - position = (1, 1, 1)
    - position = (0, 0, 1) and rotation = (45, 35.264389682, 0)
    """
    assembly = None
    if isinstance(cad_obj, (PartGroup, Part)):
        assembly = cad_obj.to_assembly()

    if assembly is None:
        raise ValueError("%s cannot be viewed" % type(cad_obj))
    return _show(assembly, **kwargs)



#
# Copyright 2019 Bernhard Walter
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from dataclasses import dataclass, field
from typing import Any, List, Dict
import warnings

from IPython.display import display
from IPython import get_ipython

from ipywidgets import HBox, Output, SelectMultiple, Layout

import cadquery as cq
from jupyter_cadquery.cadquery import Part, show
from jupyter_cadquery.cadquery.cqparts import is_cqparts_part, convert_cqparts
from jupyter_cadquery.cad_display import get_or_create_display
from .cad_objects import to_assembly
from jupyter_cadquery.cad_objects import _combined_bb
from jupyter_cadquery.defaults import get_default

#
# The Runtime part
# It hooks into __getattribute__ and intercept EVERY python call
# One should only enable replay when necessary for debugging
#


def attributes(names):
    def wrapper(cls):
        for name in names:

            def fget(self, name=name):
                if self.is_empty():
                    raise ValueError("Context empty")
                return self.stack[-1][name]

            def fset(self, value, name=name):
                if self.is_empty():
                    raise ValueError("Context empty")
                self.stack[-1][name] = value

            setattr(cls, name, property(fget, fset))
        return cls

    return wrapper


@attributes(("func", "args", "kwargs", "obj", "children"))
class Context(object):
    def __init__(self):
        self.stack = []
        self.new()

    def _to_dict(self, func, args, kwargs, obj, children):
        return {
            "func": func,
            "args": args,
            "kwargs": kwargs,
            "obj": obj,
            "children": children,
        }

    def new(self):
        self.push(None, None, None, None, [])

    def clear(self):
        self.stack = []

    def is_empty(self):
        return self.stack == []

    def is_top_level(self):
        return len(self.stack) == 1

    def pop(self):
        if not self.is_empty():
            result = self.stack[-1]
            self.stack = self.stack[:-1]
            return result
        else:
            raise ValueError("Empty context")

    def push(self, func, args, kwargs, obj, children):
        self.stack.append(self._to_dict(func, args, kwargs, obj, children))

    def append(self, func, args, kwargs, obj, children):
        self.stack[-1].append(self._to_dict(func, args, kwargs, obj, children))

    def update(self, func, args, kwargs, obj=None, children=None):
        self.func = func
        self.args = args
        self.kwargs = kwargs
        if obj is not None:
            self.obj = obj
        if children is not None:
            self.children = children

    def append_child(self, context):
        self.stack[-1]["children"].append(context)

    def __repr__(self):
        if self.is_empty():
            result = "   >> Context empty"
        else:
            result = ""
            for i, e in enumerate(self.stack):
                result += "  >> %d: %s(%s, %s); %s / %s\n" % (
                    i,
                    e["func"],
                    e["args"],
                    e["kwargs"],
                    e["obj"],
                    e["children"],
                )
        return result


_CTX = Context()
DEBUG = True
REPLAY = False


def _trace(*objs):
    if DEBUG:
        print(*objs)


def _add_context(self, name):
    def _blacklist(name):
        return name.startswith("_") or name in (
            "Workplane",
            "val",
            "vals",
            "all",
            "size",
            "add",
            "toOCC",
            "findSolid",
            "findFace",
            "toSvg",
            "exportSvg",
            "largestDimension",
        )

    def _is_recursive(func):
        return func in ["union", "cut", "intersect"]

    def intercept(parent, func):
        def f(*args, **kwargs):
            _trace("1  calling", func.__name__, args, kwargs)
            _trace(_CTX)

            if _is_recursive(func.__name__):
                _ = _CTX.pop()
                _trace("  --> level down")
                _trace(_CTX)

            if _CTX.args is None:
                _trace("2  updating")
                _CTX.update(func.__name__, args, kwargs)
                _trace(_CTX)

            result = func(*args, **kwargs)

            if func.__name__ == _CTX.func:
                _CTX.obj = result

                if _CTX.is_top_level():
                    result._caller = _CTX.pop()
                    _trace("<== _caller", func.__name__, result._caller)
                else:
                    context = _CTX.pop()
                    _CTX.append_child(context)
                    _trace("<== added child", context)
                _CTX.new()

            _trace("3  leaving", func.__name__)
            _trace(_CTX)
            return result

        return f

    attr = object.__getattribute__(self, name)
    if callable(attr):
        if not _blacklist(attr.__name__):
            _trace("==> intercepting", attr.__name__)
            if _is_recursive(attr.__name__):
                _trace("  --> level up")
                _CTX.new()
            _trace(_CTX)

            return intercept(self, attr)
    return attr


#
# The UI part
# It evaluates and optimizes what the runtime part had collected
#


@dataclass
class Step:
    level: int = 0
    func: str = ""
    args: List[Any] = field(default_factory=list)
    kwargs: Dict[Any, str] = field(default_factory=dict)
    var: str = ""
    result_name: str = ""
    result_obj: Any = None

    def clear_func(self):
        self.func = self.args = self.kwargs = ""


class Replay(object):
    def __init__(self, quality, deviation, angular_tolerance, edge_accuracy, debug, cad_width, height):
        self.debug_output = Output()
        self.quality = quality
        self.deviation = deviation
        self.angular_tolerance = angular_tolerance
        self.edge_accuracy = edge_accuracy
        self.debug = debug
        self.cad_width = cad_width
        self.height = height
        self.display = get_or_create_display()
        self.reset_camera = True

    def format_steps(self, raw_steps):
        def to_code(step, results):
            def to_name(obj):
                if isinstance(obj, cq.Workplane):
                    name = results.get(obj, None)
                else:
                    name = str(obj)
                return obj if name is None else name

            if step.func != "":
                if step.func == "newObject":
                    args = ("...",)
                else:
                    args = tuple([to_name(arg) for arg in step.args])
                code = "%s%s%s" % ("| " * step.level, step.func, args)
                code = code[:-2] if len(step.args) == 1 else code[:-1]
                if len(step.args) > 0 and len(step.kwargs) > 0:
                    code += ","
                if step.kwargs != {}:
                    code += ", ".join(["%s=%s" % (k, v) for k, v in step.kwargs.items()])
                code += ")"
                if step.result_name != "":
                    code += " => %s" % step.result_name
            elif step.var != "":
                code = "%s%s" % ("| " * step.level, step.var)
            else:
                code = "ERROR"
            return code

        steps = []
        entries = []
        obj_index = 1

        results = {step.result_obj: None for step in raw_steps}

        for i in range(len(raw_steps)):
            step = raw_steps[i]
            next_level = step.level if i == (len(raw_steps) - 1) else raw_steps[i + 1].level

            # level change, so add/use the variable name
            if step.level > 0 and step.level != next_level and step.result_name == "":
                obj_name = "_v%d" % obj_index
                obj_index += 1
                step.result_name = obj_name
            steps.append(step)

        for step in steps:
            if results[step.result_obj] is None:
                # first occurence, take note and keep
                results[step.result_obj] = step.result_name
            else:
                # next occurences remove function and add variable name
                step.var = results[step.result_obj]
                step.clear_func()

        last_level = 1000000
        for step in reversed(steps):
            if step.level < last_level:
                last_level = 1000000
                entries.insert(0, (to_code(step, results), step.result_obj))
                if step.var != "":
                    last_level = step.level

        return entries

    def to_array(self, workplane, level=0, result_name=""):
        def walk(caller, level=0, result_name=""):
            stack = [
                Step(
                    level,
                    func=caller["func"],
                    args=caller["args"],
                    kwargs=caller["kwargs"],
                    result_name=result_name,
                    result_obj=caller["obj"],
                )
            ]
            for child in reversed(caller["children"]):
                stack = walk(child, level + 1) + stack
                for arg in child["args"]:
                    if isinstance(arg, cq.Workplane):
                        result_name = getattr(arg, "name", None)
                        stack = self.to_array(arg, level=level + 2, result_name=result_name) + stack
            return stack

        stack = []

        obj = workplane
        while obj is not None:
            caller = getattr(obj, "_caller", None)
            result_name = getattr(obj, "name", "")
            if caller is not None:
                stack = walk(caller, level, result_name) + stack
                for arg in caller["args"]:
                    if isinstance(arg, cq.Workplane):
                        result_name = getattr(arg, "name", "")
                        stack = self.to_array(arg, level=level + 1, result_name=result_name) + stack
            obj = obj.parent

        return stack

    def select_handler(self, change):
        with self.debug_output:
            if change["name"] == "index":
                self.select(change["new"])

    def select(self, indexes):
        self.debug_output.clear_output()
        with self.debug_output:
            self.indexes = indexes
            cad_objs = [self.stack[i][1] for i in self.indexes]

        # Add hidden result to start with final size and allow for comparison
        if not isinstance(self.stack[-1][1].val(), cq.Vector):
            result = Part(self.stack[-1][1], "Result", show_faces=False, show_edges=False)
            objs = [result] + cad_objs
        else:
            objs = cad_objs

        with self.debug_output:
            assembly = to_assembly(*objs)
            mapping = assembly.to_state()
            shapes = assembly.collect_mapped_shapes(
                mapping,
                quality=self.quality,
                deviation=self.deviation,
                angular_tolerance=self.angular_tolerance,
                edge_accuracy=self.edge_accuracy,
                render_edges=get_default("render_edges"),
                render_normals=get_default("render_normals"),
            )
            tree = assembly.to_nav_dict()

            self.display.add_shapes(
                shapes=shapes, mapping=mapping, tree=tree, bb=_combined_bb(shapes), reset_camera=self.reset_camera
            )
            self.reset_camera = False


def replay(
    cad_obj,
    index=-1,
    quality=None,
    deviation=0.1,
    angular_tolerance=0.2,
    edge_accuracy=None,
    debug=False,
    cad_width=600,
    height=600,
):

    if not REPLAY:
        print("Replay is not enabled. To do so call 'enable_replay()'. Falling back to 'show()'")
        return show(cad_obj, cad_width=cad_width, height=height)
    else:
        print("Use the multi select box below to select one or more steps you want to examine")

    r = Replay(quality, deviation, angular_tolerance, edge_accuracy, debug, cad_width, height)

    if isinstance(cad_obj, cq.Workplane):
        workplane = cad_obj
    elif is_cqparts_part(cad_obj):
        workplane = convert_cqparts(cad_obj, replay=True)
    else:
        print("Cannot replay", cad_obj)
        return None

    r.stack = r.format_steps(r.to_array(workplane, result_name=getattr(workplane, "name", None)))
    if index == -1:
        r.indexes = [len(r.stack) - 1]
    else:
        r.indexes = [index]

    r.select_box = SelectMultiple(
        options=["[%02d] %s" % (i, code) for i, (code, obj) in enumerate(r.stack)],
        index=r.indexes,
        rows=len(r.stack),
        description="",
        disabled=False,
        layout=Layout(width="600px"),
    )
    r.select_box.add_class("monospace")
    r.select_box.observe(r.select_handler)
    display(HBox([r.select_box, r.debug_output]))

    r.select(r.indexes)
    return r


#
# Control functions to enable, disable and reset replay
#


def reset_replay():
    def warning_on_one_line(message, category, filename, lineno, file=None, line=None):
        return "%s: %s" % (category.__name__, message)

    warn_format = warnings.formatwarning
    warnings.formatwarning = warning_on_one_line
    warnings.simplefilter("always", RuntimeWarning)
    warnings.warn("jupyter_cadquery replay is enabled, turn off with disable_replay()", RuntimeWarning)
    warnings.formatwarning = warn_format

    _CTX.clear()
    _CTX.new()


def enable_replay(warning=True, debug=False):
    global DEBUG, REPLAY

    DEBUG = debug

    print("\nEnabling jupyter_cadquery replay")
    cq.Workplane.__getattribute__ = _add_context

    if warning:
        print("Note: To get rid of this warning, use 'enable_replay(False)'")
        ip = get_ipython()
        if not "reset_replay" in [f.__name__ for f in ip.events.callbacks["pre_run_cell"]]:
            ip.events.register("pre_run_cell", reset_replay)
    REPLAY = True


def disable_replay():
    global REPLAY
    print("Removing replay from cadquery.Workplane (will show a final RuntimeWarning if not suppressed)")
    cq.Workplane.__getattribute__ = object.__getattribute__

    ip = get_ipython()
    if "reset_replay" in [f.__name__ for f in ip.events.callbacks["pre_run_cell"]]:
        ip.events.unregister("pre_run_cell", reset_replay)
    REPLAY = False



import math
import numpy as np
import time
import warnings
from webcolors import name_to_rgb, hex_to_rgb, rgb_to_hex
import ipywidgets as widgets


class Color:
    def __init__(self, color=None):
        if color is None:
            self.r = self.g = self.b = 160
        elif isinstance(color, Color):
            self.r, self.g, self.b = color.r, color.g, color.b
        elif isinstance(color, str):
            if color[0] == "#":
                c = hex_to_rgb(color)
            else:
                c = name_to_rgb(color)
            self.r = c.red
            self.g = c.green
            self.b = c.blue
        elif isinstance(color, (tuple, list)) and len(color) == 3:
            if all((isinstance(c, float) and (c <= 1.0) and (c >= 0.0)) for c in color):
                self.r, self.g, self.b = (int(c * 255) for c in color)
            elif all((isinstance(c, int) and (c <= 255) and (c >= 0)) for c in color):
                self.r, self.g, self.b = color
            else:
                self._invalid(color)
        else:
            self._invalid(color)

    def __str__(self):
        return f"Color({self.r}, {self.g}, {self.b})"

    def _invalid(self, color):
        print(f"warning: {color} is an invalid color, using grey (#aaa)")
        self.r = self.g = self.b = 160

    @property
    def rgb(self):
        return (self.r, self.g, self.b)

    @property
    def percentage(self):
        return (self.r / 255, self.g / 255, self.b / 255)

    @property
    def web_color(self):
        return rgb_to_hex((self.r, self.g, self.b))


def explode(edge_list):
    return [[edge_list[i], edge_list[i + 1]] for i in range(len(edge_list) - 1)]


def flatten(nested_list):
    return [y for x in nested_list for y in x]


# CAD helpers


def distance(v1, v2):
    return np.linalg.norm([x - y for x, y in zip(v1, v2)])


def rad(deg):
    return deg / 180.0 * math.pi


def rotate_x(vector, angle):
    angle = rad(angle)
    mat = np.array(
        [
            [1, 0, 0],
            [0, math.cos(angle), -math.sin(angle)],
            [0, math.sin(angle), math.cos(angle)],
        ]
    )
    return tuple(np.matmul(mat, vector))


def rotate_y(vector, angle):
    angle = rad(angle)
    mat = np.array(
        [
            [math.cos(angle), 0, math.sin(angle)],
            [0, 1, 0],
            [-math.sin(angle), 0, math.cos(angle)],
        ]
    )
    return tuple(np.matmul(mat, vector))


def rotate_z(vector, angle):
    angle = rad(angle)
    mat = np.array(
        [
            [math.cos(angle), -math.sin(angle), 0],
            [math.sin(angle), math.cos(angle), 0],
            [0, 0, 1],
        ]
    )
    return tuple(np.matmul(mat, vector))


def rotate(vector, angle_x=0, angle_y=0, angle_z=0):
    v = tuple(vector)
    if angle_z != 0:
        v = rotate_z(v, angle_z)
    if angle_y != 0:
        v = rotate_y(v, angle_y)
    if angle_x != 0:
        v = rotate_x(v, angle_x)
    return v


def pp_vec(v):
    return "(" + ", ".join([f"{o:10.5f}" for o in v]) + ")"


def pp_loc(loc, format=True):
    T = loc.wrapped.Transformation()
    t = T.Transforms()
    q = T.GetRotation()
    if format:
        return pp_vec(t) + ", " + pp_vec((q.X(), q.Y(), q.Z(), q.W()))
    else:
        return (t, (q.X(), q.Y(), q.Z(), q.W()))


#
# tree search
#


def tree_find_single_selector(tree, selector):
    if tree.name == selector:
        return tree

    for c in tree.children:
        result = tree_find_single_selector(c, selector)
        if result is not None:
            return result
    return None


class Timer:
    def __init__(self, timeit, name, activity, level=0):
        if isinstance(timeit, bool):
            self.timeit = 99 if timeit else -1
        else:
            self.timeit = timeit
        self.activity = activity
        self.name = name
        self.level = level
        self.info = ""
        self.start = time.time()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if self.level <= self.timeit:
            prefix = ""
            if self.level > 0:
                prefix += "| " * self.level

            if self.name != "":
                self.name = f'"{self.name}"'

            print("%8.3f sec: %s%s %s %s" % (time.time() - self.start, prefix, self.activity, self.name, self.info))


class Progress:
    def __init__(self, max_, width):
        self.max = max_
        self.progress = widgets.IntProgress(
            0,
            0,
            max_,
            layout=widgets.Layout(
                width=f"{width}px", height="10px", padding="0px 4px 0px 0px !important", margin="-3px 0px -3px 2px"
            ),
        )
        self.progress.add_class("jc-progress")

    def reset(self, max_):
        self.max = max_
        self.progress.value = 0
        self.progress.max = max_

    def update(self):
        if self.progress.value < self.max:
            self.progress.value += 1


def px(w):
    return f"{w}px"


def warn(msg):
    def warning_on_one_line(message, category, filename, lineno, file=None, line=None):
        return "%s: %s" % (category.__name__, message)

    warn_format = warnings.formatwarning
    warnings.formatwarning = warning_on_one_line
    warnings.simplefilter("always", RuntimeWarning)
    warnings.warn(msg + "\n", RuntimeWarning)
    warnings.formatwarning = warn_format


#
# Copyright 2019 Bernhard Walter
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy as np
from pyquaternion import Quaternion
from pythreejs import (
    NumberKeyframeTrack,
    AnimationAction,
    AnimationClip,
    AnimationMixer,
    QuaternionKeyframeTrack,
)
from jupyter_cadquery.viewer.client import send


def _three_to_pyquat(quat):
    return Quaternion(quat[3], *quat[:3])


def _pyquat_to_three(quat):
    return (quat.x, quat.y, quat.z, quat.w)


def _from_axis_angle(axis, angle):
    return Quaternion(axis=axis, degrees=angle)


valid_transforms = ["t", "tx", "ty", "tz", "q", "rx", "ry", "rz"]


class AnimationException(BaseException):
    ...


class Animation:
    def __init__(self, root=None, viewer=False):
        if viewer and root is not None:
            print("Viewer can only animate last root, so parameter has to be None")
        elif not viewer and root is None:
            print("root group of assembly needs to be provided")

        self.viewer = viewer
        self.root = root
        self.tracks = []

    def add_track(self, selector, action, times, values):

        if self.viewer:
            self.tracks.append((selector, action, times, values))
            return

        if len(times) != len(values):
            raise AnimationException("times and values arrays need have the same lenght")

        selector = selector.replace("/", "\\")
        group = self.root.find_group(selector)
        if group is None:
            raise AnimationException(f"group '{selector}' not found")

        if action.startswith("t"):
            position = np.array(group.position).astype(np.float32)
            if action == "t":
                new_values = [position + v for v in values]
            elif action == "tx":
                new_values = [position + (v, 0, 0) for v in values]
            elif action == "ty":
                new_values = [position + (0, v, 0) for v in values]
            elif action == "tz":
                new_values = [position + (0, 0, v) for v in values]
            else:
                raise AnimationException(f"action {action} is not supported")

            self.tracks.append(
                NumberKeyframeTrack(
                    name=selector + ".position",
                    times=np.array(times).astype(np.float32),
                    values=new_values,
                )
            )

        else:
            if action.startswith("r"):
                actual = _three_to_pyquat(group.quaternion)
                if action == "rx":
                    rot_values = [_from_axis_angle([1, 0, 0], v) for v in values]
                elif action == "ry":
                    rot_values = [_from_axis_angle([0, 1, 0], v) for v in values]
                elif action == "rz":
                    rot_values = [_from_axis_angle([0, 0, 1], v) for v in values]
                else:
                    raise AnimationException(f"action {action} not supported")
                new_values = [_pyquat_to_three(actual * rot) for rot in rot_values]

            elif action == "q":
                actual = _three_to_pyquat(group.quaternion)
                new_values = [_pyquat_to_three(actual * _three_to_pyquat(q)) for q in values]

            else:
                raise AnimationException(f"action {action} is not supported")

            self.tracks.append(
                QuaternionKeyframeTrack(
                    name=selector + ".quaternion",
                    times=np.array(times).astype(np.float32),
                    values=new_values,
                )
            )

    def animate(self, speed=1, autoplay=False):
        if self.viewer:
            data = {"tracks": self.tracks, "type": "animation", "speed": speed, "autoplay": autoplay}
            send(data)
            return

        if speed != 1:
            for track in self.tracks:
                track.times = track.times / float(speed)
        clip = AnimationClip(tracks=self.tracks)
        action = AnimationAction(AnimationMixer(self.root), clip, self.root)
        if autoplay:
            action.play()
        return action



#
# Copyright 2019 Bernhard Walter
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import itertools
import math
import numpy as np
import warnings

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from pythreejs import (
        CombinedCamera,
        Plane,
        Mesh,
        LineSegments2,
        AmbientLight,
        DirectionalLight,
        Scene,
        Renderer,
        Picker,
        Group,
    )

    try:
        from orbitcontrol_patch import OrbitControls
    except:
        from pythreejs import OrbitControls


from jupyter_cadquery_widgets.widgets import state_diff
from .cad_helpers import Grid, Axes
from .utils import rotate, Color, Timer
from .cad_renderer import CadqueryRenderer, IndexedMesh
from .defaults import get_default


class CadqueryView(object):
    def __init__(
        self,
        width=600,
        height=400,
        info=None,
        timeit=False,
    ):

        self.width = width
        self.height = height
        self.info = info
        self.timeit = timeit

        self.all_shapes = None

        self.pick_color = Color("LightGreen")
        self.default_mesh_color = Color(get_default("default_color"))
        self.default_edge_color = Color(get_default("default_edgecolor"))

        self.camera_distance_factor = 6
        self.camera_initial_zoom = get_default("zoom")

        self.features = ["mesh", "edges"]

        self.bb = None

        self.pickable_objects = None
        self.pick_last_mesh = None
        self.pick_last_mesh_color = None
        self.pick_mapping = {}

        self.camera = None
        self.axes = None
        self.grid = None
        self.scene = None
        self.controller = None
        self.renderer = None

        self.position = None
        self.rotation = None
        self.zoom = None

        self.initial_rotation = None
        self.initial_position = None
        self.initial_zoom = None

        self.savestate = ((0, 0, 0, "XYZ"), (0, 0, 0))

    def get_transparent(self):
        # if one object is transparent, all are
        return self.pickable_objects.children[0].material.transparent

    def _scale(self, vec):
        r = self.bb.max_dist_from_center() * self.camera_distance_factor
        n = np.linalg.norm(vec)
        new_vec = [v / n * r for v in vec]
        return new_vec

    def _add(self, vec1, vec2):
        return list(v1 + v2 for v1, v2 in zip(vec1, vec2))

    def _sub(self, vec1, vec2):
        return list(v1 - v2 for v1, v2 in zip(vec1, vec2))

    def _norm(self, vec):
        n = np.linalg.norm(vec)
        return [v / n for v in vec]

    def _minus(self, vec):
        return [-v for v in vec]

    def direction(self):
        return self._norm(self._sub(self.camera.position, self.bb.center))

    def set_plane(self, i):
        plane = self.clippingPlanes[i]
        plane.normal = self._minus(self.direction())

    def _update(self):
        self.controller.exec_three_obj_method("update")

    def _fix_camera(self):
        zoom = self.camera.zoom
        # force the camera to zoom correctly. Seems to be a bug.
        self.camera.zoom = zoom + 1e-6
        self.camera.zoom = zoom

    def _reset_camera(self):
        self.camera.rotation, self.controller.target = self.savestate
        self.position = self.camera.position = self.initial_position
        self.zoom = self.camera.zoom = self.initial_zoom
        self._update()
        self._fix_camera()

    def _get_group(self, group_index):
        try:
            group = self.pickable_objects
            for j in group_index:
                group = group.children[j]
            return group
        except:
            return None

    def set_axes_visibility(self, value):
        self.axes.set_visibility(value)

    def set_grid_visibility(self, value):
        self.grid.set_visibility(value)

    def set_axes_center(self, value):
        self.grid.set_center(value)
        self.axes.set_center(value)

    def toggle_ortho(self, value):
        self.camera.mode = "orthographic" if value else "perspective"

    def set_transparent(self, value):
        def toggle(group, value):
            for obj in group.children:
                if isinstance(obj, Group):
                    toggle(obj, value)
                else:
                    if type(obj) is IndexedMesh:
                        obj.material.transparent = value

        toggle(self.pickable_objects, value)

    def set_black_edges(self, value):
        def toggle(group, value):
            for obj in group.children:
                if isinstance(obj, Group):
                    toggle(obj, value)
                else:
                    if isinstance(obj, LineSegments2):
                        if obj.material.linewidth == 1:
                            obj.material.color = "#000" if value else self.edge_color

        toggle(self.pickable_objects, value)

    def set_visibility(self, ind, i, state):
        feature = self.features[i]
        group_index = self.pick_mapping[ind][feature]
        group = self._get_group(group_index)
        if group is not None:
            group.visible = state == 1

    def change_visibility(self, paths):
        def f(states):
            diffs = state_diff(states.get("old"), states.get("new"))
            for diff in diffs:
                [[obj, val]] = diff.items()
                self.set_visibility(paths[obj], val["icon"], val["new"])

        return f

    def set_clipping(self, tab):
        if tab == 0:
            self.renderer.clippingPlanes = []
        else:
            self.renderer.clippingPlanes = self.clippingPlanes

    def _filter_shapes(self, shapes):
        return {
            "parts": [
                (
                    {"id": shape["id"], "name": shape["name"], "bb": shape["bb"]}
                    if shape.get("parts") is None
                    else self._filter_shapes(shape)
                )
                for shape in shapes["parts"]
            ]
        }

    def _get_bb(self, shape_index):
        shape = self.bbs
        try:
            for j in shape_index:
                shape = shape["parts"][j]
        except:
            return None
        return shape

    def pick(self, value):
        if self.pick_last_mesh != value.owner.object:
            # Reset
            if value.owner.object is None or self.pick_last_mesh is not None:
                self.pick_last_mesh.material.color = self.pick_last_mesh_color
                self.pick_last_mesh = None
                self.pick_last_mesh_color = None

            # Change highlighted mesh
            if isinstance(value.owner.object, Mesh):
                self.pick_last_mesh = value.owner.object
                shape = self._get_bb(value.owner.object.ind["shape"])
                bbox = shape["bb"]

                self.info.bb_info(
                    shape["name"],
                    (
                        (bbox["xmin"], bbox["xmax"]),
                        (bbox["ymin"], bbox["ymax"]),
                        (bbox["zmin"], bbox["zmax"]),
                        (
                            (bbox["xmax"] + bbox["xmin"]) / 2,
                            (bbox["ymax"] + bbox["ymin"]) / 2,
                            (bbox["zmax"] + bbox["zmin"]) / 2,
                        ),
                    ),
                )
                self.pick_last_mesh_color = self.pick_last_mesh.material.color
                self.pick_last_mesh.material.color = self.pick_color.web_color

    def clip(self, index):
        def f(change):
            self.clippingPlanes[index].constant = change["new"]

        return f

    # public methods to render the view

    def is_ortho(self):
        return self.camera.mode == "orthographic"

    def is_transparent(self):
        return self.pickable_objects.children[0].material.transparent

    def create(self):
        self.cq_renderer = CadqueryRenderer(
            default_mesh_color=self.default_mesh_color,
            default_edge_color=self.default_edge_color,
            timeit=self.timeit,
        )

        # Set up camera
        self.camera = CombinedCamera(
            position=(1.0, 1.0, 1.0),
            width=self.width,
            height=self.height,
            far=100,
            orthoFar=100,
            up=(0.0, 0.0, 1.0),
        )
        # needs to be an extra step to take effect
        self.toggle_ortho(True)

        # Set up scene
        self.scene = Scene(
            children=[
                self.camera,
            ]
        )
        if get_default("theme") == "dark":
            self.scene.background = "#212121"

        # Set up Controllers
        camera_target = (0.0, 0.0, 0.0)
        self.controller = OrbitControls(controlling=self.camera, target=camera_target, target0=camera_target)

        # Create Renderer instance
        self.renderer = Renderer(
            scene=self.scene,
            camera=self.camera,
            controls=[self.controller],
            antialias=True,
            width=self.width,
            height=self.height,
        )
        return self.renderer

    def add_shapes(
        self,
        shapes,
        bb,
        ticks,
        progress,
        bb_factor=None,
        ambient_intensity=None,
        direct_intensity=None,
        default_edgecolor=None,
        position=None,
        rotation=None,
        zoom=None,
        reset_camera=True,
    ):

        preset = lambda key, value: get_default(key) if value is None else value

        bb_factor = preset("bb_factor", bb_factor)
        ticks = preset("ticks", ticks)
        ambient_intensity = preset("ambient_intensity", ambient_intensity)
        direct_intensity = preset("direct_intensity", direct_intensity)
        self.edge_color = Color(preset("default_edgecolor", default_edgecolor)).web_color
        self.cq_renderer.default_edge_color = self.edge_color

        self.bbs = self._filter_shapes(shapes)
        self.bb = bb

        # Render Shapes
        with Timer(self.timeit, "", "overall render", 3):
            self.pickable_objects, self.pick_mapping = self.cq_renderer.render(shapes, progress)

        with Timer(self.timeit, "", "configure view", 3):
            bb_max = self.bb.max_dist_from_center()
            orbit_radius = 4 * bb_factor * bb_max

            if reset_camera:
                self.initial_rotation = self.rotation = preset("rotation", rotation)
                self.initial_position = self.position = self._add(
                    self.bb.center, self._scale(rotate(preset("position", position), *self.rotation))
                )
                self.initial_zoom = self.zoom = preset("zoom", zoom)
            else:
                if rotation is not None:
                    self.rotation = rotation
                if self.rotation is None:
                    self.rotation = get_default("rotation")
                if self.initial_rotation is None:
                    self.initial_rotation = self.rotation

                if position is not None:
                    self.position = position
                if self.position is None:
                    self.position = self._add(
                        self.bb.center, self._scale(rotate(preset("position", position), *self.rotation))
                    )
                if self.initial_position is None:
                    self.initial_position = self.position

                if zoom is not None:
                    self.zoom = zoom
                if self.zoom is None:
                    self.zoom = get_default("zoom")
                if self.initial_zoom is None:
                    self.initial_zoom = self.zoom

            # Set up Helpers relative to bounding box
            xy_max = max(abs(self.bb.xmin), abs(self.bb.xmax), abs(self.bb.ymin), abs(self.bb.ymax)) * 1.2
            self.grid = Grid(
                bb_center=self.bb.center, maximum=xy_max, colorCenterLine="#aaa", colorGrid="#ddd", ticks=ticks
            )
            self.grid.set_visibility(False)

            self.axes = Axes(bb_center=self.bb.center, length=self.grid.grid.size / 2)
            self.axes.set_visibility(False)

            # Set up the controller relative to bounding box
            self.controller.target = self.bb.center
            self.controller.target0 = self.bb.center
            self.controller.panSpeed = (self.bb.xsize + self.bb.ysize + self.bb.zsize) / 300

            # Set up lights in every of the 8 corners of the global bounding box
            positions = list(itertools.product(*[(-orbit_radius, orbit_radius)] * 3))

            self.amb_light = AmbientLight(intensity=ambient_intensity)
            self.key_lights = [
                DirectionalLight(color="white", position=position, intensity=direct_intensity)
                for position in positions
            ]

            # Set up Picker
            self.picker = Picker(controlling=self.pickable_objects, event="dblclick")
            self.picker.observe(self.pick)
            self.renderer.controls = self.renderer.controls + [self.picker]

            # Set up camera
            self.update_camera(self.position, self.zoom, orbit_radius)

            # Add objects to scene
            self._fix_camera()
            self.add_to_scene()

            # needs to be done after setup of camera
            self.grid.set_rotation((math.pi / 2.0, 0, 0, "XYZ"))
            self.grid.set_position((0, 0, 0))

            self.renderer.localClippingEnabled = True
            self.renderer.clippingPlanes = []  # turn off when not in clipping view
            self.clippingPlanes = [
                Plane((-1, 0, 0), 0.02 if abs(self.bb.xmax) < 1e-4 else self.bb.xmax * bb_factor),
                Plane((0, -1, 0), 0.02 if abs(self.bb.ymax) < 1e-4 else self.bb.ymax * bb_factor),
                Plane((0, 0, -1), 0.02 if abs(self.bb.zmax) < 1e-4 else self.bb.zmax * bb_factor),
            ]

            self.savestate = (self.camera.rotation, self.controller.target)

        return self.renderer

    def update_camera(self, position, zoom, orbit_radius):
        self.camera.position = position
        self.camera.zoom = zoom
        self.camera.far = 10 * orbit_radius
        self.camera.orthoFar = 10 * orbit_radius
        self._update()

    def add_to_scene(self):
        self.scene.add(self.amb_light)
        self.scene.add(self.key_lights)
        self.scene.add(self.axes.axes)
        self.scene.add(self.grid.grid)
        self.scene.add(self.pickable_objects)

    def clear(self):
        # save camera position and zoom in case we want to keep it for the object
        self.zoom = self.camera.zoom
        self.position = self.camera.position

        self.scene.remove(self.amb_light)
        self.scene.remove(self.key_lights)
        self.scene.remove(self.axes.axes)
        self.scene.remove(self.grid.grid)
        self.scene.remove(self.pickable_objects)

    @property
    def root_group(self):
        return self.pickable_objects



#
# Copyright 2021 Bernhard Walter
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


class Defaults:
    def __init__(self):
        self.reset_defaults()

    def get_defaults(self):
        return self.defaults

    def get_default(self, key, default_value=None):
        return self.defaults.get(key, default_value)

    def set_defaults(self, **kwargs):
        """Set defaults for CAD viewer

        Valid keywords:
        - height:            Height of the CAD view (default=600)
        - tree_width:        Width of navigation tree part of the view (default=250)
        - cad_width:         Width of CAD view part of the view (default=800)
        - bb_factor:         Scale bounding box to ensure compete rendering (default=1.5)
        - default_color:     Default mesh color (default=(232, 176, 36))
        - default_edgecolor: Default mesh color (default=(128, 128, 128))
        - render_edges:      Render edges  (default=True)
        - render_normals:    Render normals (default=False)
        - render_mates:      Render mates (for MAssemblies)
        - mate_scale:        Scale of rendered mates (for MAssemblies)
        - quality:           Linear deflection for tessellation (default=None)
                             If None, uses: (xlen + ylen + zlen) / 300 * deviation)
        - deviation:         Deviation from default for linear deflection value ((default=0.1)
        - angular_tolerance: Angular deflection in radians for tessellation (default=0.2)
        - edge_accuracy:     Presicion of edge discretizaion (default=None)
                             If None, uses: quality / 100
        - optimal_bb:        Use optimal bounding box (default=False)
        - axes:              Show axes (default=False)
        - axes0:             Show axes at (0,0,0) (default=False)
        - grid:              Show grid (default=False)
        - ticks:             Hint for the number of ticks in both directions (default=10)
        - ortho:             Use orthographic projections (default=True)
        - transparent:       Show objects transparent (default=False)
        - ambient_intensity  Intensity of ambient ligth (default=1.0)
        - direct_intensity   Intensity of direct lights (default=0.12)
        - position:          Relative camera position that will be scaled (default=(1, 1, 1))
        - rotation:          z, y and y rotation angles to apply to position vector (default=(0, 0, 0))
        - zoom:              Zoom factor of view (default=2.5)
        - reset_camera:      Reset camera position, rotation and zoom to default (default=True)
        - mac_scrollbar:     Prettify scrollbars (default=True)
        - display:           Select display: "sidecar", "cell", "html"
        - theme:             Theme "light" or "dark" (default="light")
        - tools:             Show the viewer tools like the object tree
        - timeit:            Show rendering times, levels = False, 0,1,2,3,4,5 (default=False)

        For example isometric projection can be achieved in two ways:
        - position = (1, 1, 1)
        - position = (0, 0, 1) and rotation = (45, 35.264389682, 0)
        """

        for k, v in kwargs.items():
            if self.get_default(k, "") == "":
                print("Paramater %s is not a valid argument for show()" % k)
            else:
                if k == "zoom" and v == 1.0:
                    # for zoom == 1 viewing has a bug, so slightly increase it
                    v = 1 + 1e-6
                self.defaults[k] = v

        if kwargs.get("display") == "html":
            self.defaults["tools"] = False
            from IPython.display import HTML, display
            from ipywidgets.embed import DEFAULT_EMBED_REQUIREJS_URL

            display(HTML(f"""<script src="{DEFAULT_EMBED_REQUIREJS_URL}" crossorigin="anonymous"></script>"""))

    def reset_defaults(self):
        self.defaults = {
            "height": 600,
            "tree_width": 250,
            "cad_width": 800,
            "bb_factor": 1.0,
            "default_color": (232, 176, 36),
            "default_edgecolor": (128, 128, 128),
            "render_edges": True,
            "render_normals": False,
            "render_mates": False,
            "mate_scale": 1,
            "quality": None,
            "deviation": 0.1,
            "angular_tolerance": 0.2,
            "edge_accuracy": None,
            "optimal_bb": False,
            "axes": False,
            "axes0": False,
            "grid": False,
            "ticks": 10,
            "ortho": True,
            "transparent": False,
            "black_edges": False,
            "ambient_intensity": 1.0,
            "direct_intensity": 0.12,
            "position": (1, 1, 1),
            "rotation": (0, 0, 0),
            "zoom": 2.5,
            "reset_camera": True,
            "mac_scrollbar": True,
            "display": "cell",
            "theme": "light",
            "tools": True,
            "timeit": False,
        }


def get_defaults():
    return DEFAULTS.get_defaults()


def get_default(key, default_value=None):
    return DEFAULTS.get_default(key, default_value)


def set_defaults(**kwargs):
    DEFAULTS.set_defaults(**kwargs)


def reset_defaults():
    DEFAULTS.reset_defaults()


def split_args(config):
    create_args = {
        k: v
        for k, v in config.items()
        if k
        in [
            "height",
            "bb_factor",
            "tree_width",
            "cad_width",
            "axes",
            "axes0",
            "grid",
            "ortho",
            "transparent",
            "mac_scrollbar",
            "display",
            "tools",
            "timeit",
        ]
    }
    add_shape_args = {
        k: v
        for k, v in config.items()
        if k
        in [
            "bb_factor",
            "ticks",
            "default_edgecolor",
            "ambient_intensity",
            "direct_intensity",
            "position",
            "rotation",
            "zoom",
            "reset_camera",
        ]
    }
    return create_args, add_shape_args


DEFAULTS = Defaults()


from __future__ import print_function
from setuptools import setup, find_packages
from glob import glob
import os

here = os.path.dirname(os.path.abspath(__file__))
is_repo = os.path.exists(os.path.join(here, ".git"))

from distutils import log

log.set_verbosity(log.DEBUG)
log.info("setup.py entered")
log.info("$PATH=%s", os.environ["PATH"])

LONG_DESCRIPTION = "An extension to render cadquery objects in JupyterLab via pythreejs"

setup_args = {
    "name": "jupyter_cadquery",
    "version": "2.2.1",
    "description": "An extension to render cadquery objects in JupyterLab via pythreejs",
    "long_description": LONG_DESCRIPTION,
    "include_package_data": True,
    "python_requires": ">=3.6",
    "install_requires": [
        "ipywidgets~=7.6",
        "webcolors~=1.11",
        "notebook~=6.3",
        "sidecar~=0.5",
        "jupyter-cadquery-widgets~=2.0",
        "pythreejs~=2.3",
        "voila~=0.2",
        "cadquery_massembly~=0.9",
        "orbitcontrol-patch~=0.1.0",
        "pyquaternion~=0.9.9",
    ],
    "extras_require": {
        "dev": {"jupyter-packaging", "cookiecutter", "twine", "bumpversion", "black", "pylint", "pyYaml"},
        "prod": {"cadquery==2.1"},
    },
    "packages": find_packages(),
    "scripts": ["jcv", "jcv.cmd"],
    "zip_safe": False,
    "author": "Bernhard Walter",
    "author_email": "b_walter@arcor.de",
    "url": "https://github.com/bernhard-42/jupyter-cadquery",
    "keywords": [
        "ipython",
        "jupyter",
        "widgets",
    ],
    "classifiers": [
        "Development Status :: 4 - Beta",
        "Framework :: IPython",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Multimedia :: Graphics",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
}

setup(**setup_args)



#
# Copyright 2019 Bernhard Walter
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from IPython.display import display, HTML

LATEST = None

base_css = """
.scroll-area {
    overflow: scroll !important;
    border: unset !important;
}

.mac-scrollbar::-webkit-scrollbar {
    width: 5px !important;
    height: 5px !important;
}

.mac-scrollbar::-webkit-scrollbar-track {
    background-color: transparent !important;
}

.mac-scrollbar .widget-html-content {
    overflow-x: visible;
    overflow-y: visible;
}

.tab-content-no-padding .widget-tab-contents {
    overflow-x: visible !important;
    overflow-y: visible !important;
    padding-bottom: 0px !important;
}

.view_renderer {
    border: 1px solid var(--jp-border-color1);
    margin-top: 3px;
    margin-left: 2px;
}

.view_tree {
    padding: 0px !important;
}

.view_axes {
    width: 60px !important;
    margin-left: 5px !important;
}

.view_zero {
    width: 55px !important;
}

.view_grid {
    width: 56px !important;
}

.view_ortho {
    width: 64px !important;
}

.view_transparent {
    width: 125px !important;
}

.view_black_edges {
    width: 105px !important;
}

.view_button {
    padding: 0px !important;
}

.view_button>img {
    height: 28px;
    width: 36px;
}

.node_entry_wrap {
    white-space: pre;
}

.node_entry {
    white-space: nowrap;
    padding-top: 4px;
}

.t-caret {
    cursor: pointer;
    -webkit-user-select: none;
    /* Safari 3.1+ */
    -moz-user-select: none;
    /* Firefox 2+ */
    -ms-user-select: none;
    /* IE 10+ */
    user-select: none;
}

.t-caret-down::before {
    -ms-transform: rotate(90deg);
    /* IE 9 */
    -webkit-transform: rotate(90deg);
    /* Safari */
    transform: rotate(90deg);
}

.toplevel {
    list-style-type: none;
    padding-inline-start: 0px;
}

.nested {
    display: none;
    list-style-type: none;
    padding-inline-start: 16px;
}

.active {
    display: block;
}

.icon {
    width: 28px !important;
    height: 22px !important;
    padding-right: 2px;
    vertical-align: middle;
}

.indent {
    margin-left: 12px;
}

.tree_label {
    padding-left: 2px;
    font-size: 14px;
}

.scroll_down {
    display: flex;
    flex-direction: column-reverse;
}

.small_table {
    line-height: 14px;
}

.monospace select {
    font-family: monospace;
}
"""

css = {
    "light": base_css
    + """
        .t-caret::before {
            content: u"\u25B6";
            font-size: 12px;
            color: "#080808";
            display: inline-block;
            margin-right: 2px;
        }

        .mac-scrollbar::-webkit-scrollbar-thumb {
            background-color: rgba(0, 0, 0, 0.2) !important;
            border-radius: 100px !important;
        }

        .mac-scrollbar::-webkit-scrollbar-thumb:hover {
            background: rgba(0, 0, 0, 0.4) !important;
        }

        .mac-scrollbar::-webkit-scrollbar-thumb:active {
            background: #181818 !important;
        }

        .mac-scrollbar::-webkit-scrollbar-corner {
            background: white;
        }

        .view_output {
            border: 1px solid var(--jp-border-color1);
            margin: 2px 2px 2px 2px !important;
            padding-right: 1px !important;
            background-color: white;
        }
    """,
    "dark": base_css
    + """
        .t-caret::before {
            content: u"\u25B6";
            font-size: 12px;
            color: #e0e0e0;
            display: inline-block;
            margin-right: 2px;
        }

        .mac-scrollbar::-webkit-scrollbar-thumb {
            background-color: rgba(255, 255, 255, 0.3) !important;
            border-radius: 100px !important;
        }

        .mac-scrollbar::-webkit-scrollbar-thumb:hover {
            background: rgba(255, 255, 255, 0.5) !important;
        }

        .mac-scrollbar::-webkit-scrollbar-thumb:active {
            background: #e0e0e0 !important;
        }

        .mac-scrollbar::-webkit-scrollbar-corner {
            background: #212121;
        }

        .view_output {
            border: 1px solid var(--jp-border-color1);
            margin: 2px 2px 2px 2px !important;
            padding-right: 1px !important;
            background-color: #212121;
        }
    """,
}


def set_css(theme, force=False):
    global LATEST

    if force or theme != LATEST:
        display(HTML(f"""<style>{css[theme]}</style>"""))
        LATEST = theme
